


' ---------------------------------- functionality for generating tick events with pulses per quarter note resolution
'
Public MustInherit Class PpqnClock
    Implements IClock

    ' ------------------------------------------------------------ the default tempo in microseconds: 120bpm
    Public Const DefaultTempo As Integer = 500000
    ' ------------------------------------------------------------ the minimum pulses per quarter note value
    Public Const PpqnMinValue As Integer = 24
    ' ------------------------------------------------------------ the number of microseconds per millisecond
    Private Const MicrosecondsPerMillisecond As Integer = 1000
    ' ------------------------------------------------------------ the pulses per quarter note value
    Private m_ppqn As Integer = PpqnMinValue
    ' ------------------------------------------------------------ the tempo in microseconds
    Private m_tempo As Integer = DefaultTempo
    ' ------------------------------------------------------------ the product of the timer period, 
    '                                                               the pulses per quarter note, and
    '                                                               the number of microseconds per millisecond
    Private periodResolution As Integer
    ' ------------------------------------------------------------ the number of ticks per MIDI clock
    Private m_ticksPerClock As Integer
    ' ------------------------------------------------------------ the running fractional tick count
    Private fractionalTicks As Integer = 0
    ' ------------------------------------------------------------ the timer period
    Private ReadOnly timerPeriod As Integer
    ' ------------------------------------------------------------ indicates the clock running
    Protected running As Boolean = False


    Protected Sub New(ByVal timerPeriod As Integer)
        If timerPeriod < 1 Then
            Throw New ArgumentOutOfRangeException("timerPeriod", timerPeriod, "Timer period cannot be less than one.")
        End If

        Me.timerPeriod = timerPeriod

        CalculatePeriodResolution()
        CalculateTicksPerClock()
    End Sub


    ' ------------------------------------------------------------------------- "Methods"

    Protected Function GetTempo() As Integer
        Return m_tempo
    End Function

    Protected Sub SetTempo(ByVal _tempo As Integer)
        If _tempo < 1 Then
            Throw New ArgumentOutOfRangeException("Tempo out of range.")
        End If
        m_tempo = _tempo
    End Sub

    Protected Sub Reset()
        fractionalTicks = 0
    End Sub

    Protected Function GenerateTicks() As Integer
        Dim ticks As Integer = (fractionalTicks + periodResolution) \ m_tempo
        fractionalTicks += periodResolution - ticks * m_tempo
        Return ticks
    End Function

    Private Sub CalculatePeriodResolution()
        periodResolution = m_ppqn * timerPeriod * MicrosecondsPerMillisecond
    End Sub

    Private Sub CalculateTicksPerClock()
        m_ticksPerClock = m_ppqn \ PpqnMinValue
    End Sub

    Protected Overridable Sub OnTick(ByVal e As EventArgs)
        RaiseEvent Tick(Me, EventArgs.Empty)
    End Sub

    Protected Overridable Sub OnStarted(ByVal e As EventArgs)
        RaiseEvent Started(Me, e)
    End Sub

    Protected Overridable Sub OnStopped(ByVal e As EventArgs)
        RaiseEvent Stopped(Me, e)
    End Sub

    Protected Overridable Sub OnContinued(ByVal e As EventArgs)
        RaiseEvent Continued(Me, e)
    End Sub


    ' ------------------------------------------------------------------------- "Properties"

    Public Property Ppqn() As Integer
        Get
            Return m_ppqn
        End Get
        Set(ByVal value As Integer)
            If value < PpqnMinValue Then
                Throw New ArgumentOutOfRangeException("Ppqn", value, "Pulses per quarter note out of range.")
            ElseIf value Mod PpqnMinValue <> 0 Then
                Throw New ArgumentException("Pulses per quarter note is not a multiple of 24.")
            End If

            m_ppqn = value

            CalculatePeriodResolution()
            CalculateTicksPerClock()
        End Set
    End Property

    Public MustOverride ReadOnly Property Ticks() As Integer Implements IClock.Ticks

    Public ReadOnly Property TicksPerClock() As Integer
        Get
            Return m_ticksPerClock
        End Get
    End Property

    Public ReadOnly Property GetPeriodResolution() As Integer
        Get
            Return periodResolution
        End Get
    End Property


    ' ------------------------------------------------------------------------------------  "IClock Members"

    Public Event Tick As System.EventHandler Implements IClock.Tick
    Public Event Started As System.EventHandler Implements IClock.Started
    Public Event Continued As System.EventHandler Implements IClock.Continued
    Public Event Stopped As System.EventHandler Implements IClock.Stopped

    'Public Event Continued1(ByVal sender As Object, ByVal e As System.EventArgs) Implements IClock.Continued
    'Public Event Started1(ByVal sender As Object, ByVal e As System.EventArgs) Implements IClock.Started
    'Public Event Stopped1(ByVal sender As Object, ByVal e As System.EventArgs) Implements IClock.Stopped
    'Public Event Tick1(ByVal sender As Object, ByVal e As System.EventArgs) Implements IClock.Tick

    Public ReadOnly Property IsRunning() As Boolean Implements IClock.IsRunning
        Get
            Return running
        End Get
    End Property


End Class
