

Imports System.ComponentModel
Imports System.Diagnostics

' ----------------------------------------------------------- constants for ChannelMessage types.
Public Enum ChannelCommand
    NoteOff = &H80
    NoteOn = &H90
    PolyPressure = &HA0
    Controller = &HB0
    ProgramChange = &HC0
    ChannelPressure = &HD0      ' (aftertouch)
    PitchWheel = &HE0
End Enum

Public Enum ControllerType
    BankSelect
    ModulationWheel
    BreathControl
    FootPedal = 4
    PortamentoTime
    DataEntrySlider
    Volume
    Balance
    Pan = 10
    Expression
    EffectControl1
    EffectControl2
    GeneralPurposeSlider1 = 16
    GeneralPurposeSlider2
    GeneralPurposeSlider3
    GeneralPurposeSlider4
    BankSelectFine = 32
    ModulationWheelFine
    BreathControlFine
    FootPedalFine = 36
    PortamentoTimeFine
    DataEntrySliderFine
    VolumeFine
    BalanceFine
    PanFine = 42
    ExpressionFine
    EffectControl1Fine
    EffectControl2Fine
    HoldPedal1 = 64
    Portamento
    SustenutoPedal
    SoftPedal
    LegatoPedal
    HoldPedal2
    SoundVariation
    SoundTimbre
    SoundReleaseTime
    SoundAttackTime
    SoundBrightness
    SoundControl6
    SoundControl7
    SoundControl8
    SoundControl9
    SoundControl10
    GeneralPurposeButton1
    GeneralPurposeButton2
    GeneralPurposeButton3
    GeneralPurposeButton4
    EffectsLevel = 91
    TremeloLevel
    ChorusLevel
    CelesteLevel
    PhaserLevel
    DataButtonIncrement
    DataButtonDecrement
    NonRegisteredParameterFine
    NonRegisteredParameterCoarse
    RegisteredParameterFine
    RegisteredParameterCoarse
    AllSoundOff = 120
    AllControllersOff
    LocalKeyboard
    AllNotesOff
    OmniModeOff
    OmniModeOn
    MonoOperation
    PolyOperation
End Enum

' --------------------------------------------------------------------- Represents MIDI channel messages
<ImmutableObject(True)> _
Public NotInheritable Class ChannelMessage
    Inherits ShortMessage

    ' ----------------------------------------------------------------- bit manipulation constants
    Private Const MidiChannelMask As Integer = Not 15
    Private Const CommandMask As Integer = Not 240
    ' ----------------------------------------------------------------- maximum value allowed for MIDI channels
    Public Const MidiChannelMaxValue As Integer = 15


    Public Sub New(ByVal _command As ChannelCommand, _
                   ByVal _midiChannel As Integer, _
                   ByVal _data1 As Integer)
        msg = 0

        msg = PackCommand(msg, _command)
        msg = PackMidiChannel(msg, _midiChannel)
        msg = PackData1(msg, _data1)

        Debug.Assert(Command = _command)
        Debug.Assert(MidiChannel = _midiChannel)
        Debug.Assert(Data1 = _data1)
    End Sub


    Public Sub New(ByVal _command As ChannelCommand, _
                   ByVal _midiChannel As Integer, _
                   ByVal _data1 As Integer, _
                   ByVal _data2 As Integer)

        msg = 0

        msg = PackCommand(msg, _command)
        msg = PackMidiChannel(msg, _midiChannel)
        msg = PackData1(msg, _data1)
        msg = PackData2(msg, _data2)

        Debug.Assert(Command = _command)
        Debug.Assert(MidiChannel = _midiChannel)
        Debug.Assert(Data1 = _data1)
        Debug.Assert(Data2 = _data2)
    End Sub

    Friend Sub New(ByVal message As Integer)
        Me.msg = message
    End Sub

    ' ================================================================================
    '  Methods
    ' ================================================================================
    Public Overloads Overrides Function GetHashCode() As Integer
        Return msg
    End Function

    ' ------------------------------------------------ determines if two ChannelEventArgs instances are equal
    '
    Public Overloads Overrides Function Equals(ByVal obj As Object) As Boolean
        If Not (TypeOf obj Is ChannelMessage) Then
            Return False
        End If
        Dim e As ChannelMessage = DirectCast(obj, ChannelMessage)
        Return Me.msg = e.msg
    End Function

    ' ------------------------------------------------ how many bytes are used for the specified ChannelCommand.
    '
    Friend Shared Function DataBytesPerType(ByVal command As ChannelCommand) As Integer
        Dim result As Integer
        If command = ChannelCommand.ChannelPressure OrElse command = ChannelCommand.ProgramChange Then
            result = 1
        Else
            result = 2
        End If
        Return result
    End Function

    ' ------------------------------------------------ unpacks the command value from the specified integer message
    '
    Friend Shared Function UnpackCommand(ByVal message As Integer) As ChannelCommand
        Return CType((message And DataMask And MidiChannelMask), ChannelCommand)
    End Function

    ' ------------------------------------------------ unpacks the MIDI channel from the specified integer message
    '
    Friend Shared Function UnpackMidiChannel(ByVal message As Integer) As Integer
        Return message And DataMask And CommandMask
    End Function

    ' ------------------------------------------------ packs the MIDI channel into the specified integer message
    '
    Friend Shared Function PackMidiChannel(ByVal message As Integer, ByVal midiChannel As Integer) As Integer
        If midiChannel < 0 OrElse midiChannel > MidiChannelMaxValue Then
            Throw New ArgumentOutOfRangeException("midiChannel", midiChannel, "MIDI channel out of range.")
        End If
        Return (message And MidiChannelMask) Or midiChannel
    End Function

    ' ------------------------------------------------- packs the command value into an integer message
    '
    Friend Shared Function PackCommand(ByVal message As Integer, ByVal command As ChannelCommand) As Integer
        Return (message And CommandMask) Or command
    End Function


    ' ================================================================================
    '  Properties
    ' ================================================================================

    ' ---------------------------------------------------- gets or sets the channel command value
    Public Property Command() As ChannelCommand
        Get
            Return UnpackCommand(msg)
        End Get
        Set(ByVal value As ChannelCommand)
            msg = PackCommand(msg, value)
        End Set
    End Property

    ' ---------------------------------------------------- gets or sets the MIDI channel
    Public Property MidiChannel() As Integer
        Get
            Return UnpackMidiChannel(msg)
        End Get
        Set(ByVal value As Integer)
            msg = PackMidiChannel(msg, value)
        End Set
    End Property

    ' ---------------------------------------------------- gets or sets the first data value
    Public Property Data1() As Integer
        Get
            Return UnpackData1(msg)
        End Get
        Set(ByVal value As Integer)
            msg = PackData1(msg, value)
        End Set
    End Property

    ' ---------------------------------------------------- gets or sets the second data value
    Public Property Data2() As Integer
        Get
            Return UnpackData2(msg)
        End Get
        Set(ByVal value As Integer)
            msg = PackData2(msg, value)
        End Set
    End Property

    ' ---------------------------------------------------- gets the EventType
    '
    Public Overloads Overrides ReadOnly Property MessageType() As MessageType
        Get
            Return MessageType.Channel
        End Get
    End Property

End Class
