

Imports System.Diagnostics

' -------------------------------------------------------------------- a traditional ADSR envelope
Public Class AdsrEnvelope
	Inherits MonoSynthComponent
    Implements IProgrammable

    Public Enum ParameterId
        AttackTime
        DecayTime
        SustainLevel
        ReleaseTime
        TriggerMode
        VelocitySensitivity
    End Enum

    Private Enum State
        Attack
        Decay
        Release
        Completed
    End Enum

    Private Enum TriggerMode
        Retrigger
        Legato
    End Enum

    ' ---------------------------------------------------- scaler value to ensure output is in the range of [0, 1]
    Private Const Ceiling As Single = 0.63212F
    ' ---------------------------------------------------- scaler to scale up time parameters.
    Private Const TimeScaler As Integer = 3

    Private m_triggerMode As TriggerMode = TriggerMode.Retrigger
    Private attackTime As Single = 0
    Private decayTime As Single = 0.25F
    Private sustainLevel As Single
    Private releaseTime As Single = 0.25F
    Private velocitySensitivity As Single = 0

    ' ---------------------------------------------------- coefficients for calculating the AdsrEnvelope's output
    Private a0 As Single, b1 As Single
    ' ---------------------------------------------------- the AdsrEnvelope's current output
    Private output As Single = 1
    ' ---------------------------------------------------- the overall amplitude
    Private amplitude As Single = 1
    ' ---------------------------------------------------- the current state
    Private m_state As State = State.Completed
    ' ---------------------------------------------------- indicates if the AdsrEnvelope overwrit its buffer
    Private m_synthesizeReplaceEnabled As Boolean = True


    Public Sub New(ByVal sampleRate As SampleRate, ByVal buffer As MonoBuffer)
        MyBase.New(sampleRate, buffer)
        SetParameterValue(ParameterId.SustainLevel, 0)
    End Sub

    Public Sub New(ByVal sampleRate As SampleRate, ByVal buffer As MonoBuffer, ByVal name As String)
        MyBase.New(sampleRate, buffer, name)
        SetParameterValue(ParameterId.SustainLevel, 0)
    End Sub


    Public Overloads Overrides Sub Synthesize(ByVal offset As Integer, ByVal count As Integer)
        Dim buffer As Single() = GetBuffer()
        Dim c As Integer = offset + count
        Dim o As Single

        Debug.Assert(buffer.Length >= offset + count)

        For i As Integer = offset To c - 1
            Select Case m_state
                Case State.Attack
                    output = a0 + b1 * output

                    ' -------------------------------------------- the end of the attack segment has been reached
                    If output >= Ceiling + 1 Then
                        ' ---------------------------------------- calculate coefficients for decay segment
                        '
                        Dim d As Single = decayTime * TimeScaler * SamplesPerSecond + 1
                        Dim x As Single = CSng(Math.Exp(-1 / d))
                        '
                        a0 = 1 - x
                        b1 = x
                        '
                        output = Ceiling + 1
                        ' ----------------------------------------- enter decay segment
                        m_state = State.Decay
                    End If

                Case State.Decay
                    output = a0 * sustainLevel + b1 * output

                Case State.Release
                    output = a0 + b1 * output
                    ' --------------------------------------------- the end of the release segment has been reached
                    If output < 1 Then
                        output = 1
                        m_state = State.Completed
                    End If

                Case State.Completed

                Case Else
                    Debug.Fail("Unhandled state")
            End Select

            o = (output - 1) / Ceiling * amplitude

            buffer(i) = (If(m_synthesizeReplaceEnabled, o, buffer(i) + o))
        Next
    End Sub


    Public Overloads Overrides Sub Trigger(ByVal previousNote As Integer, _
                                           ByVal note As Integer, _
                                           ByVal velocity As Single)

        If m_triggerMode = TriggerMode.Legato AndAlso (m_state = State.Attack OrElse m_state = State.Decay) Then
            Return
        End If
        ' -------------------------------------------------------- calculate coefficients for the attack segment
        '
        Dim d As Single = attackTime * TimeScaler * SamplesPerSecond + 1
        Dim x As Single = CSng(Math.Exp(-1 / (attackTime * TimeScaler * SamplesPerSecond)))
        a0 = (1 - x) * 2
        b1 = x
        amplitude = CSng(Math.Pow((1 - velocitySensitivity) + velocity * velocitySensitivity, 2))
        m_state = State.Attack
    End Sub


    Public Overloads Overrides Sub Release(ByVal velocity As Single)
        If m_state = State.Release OrElse m_state = State.Completed Then
            Return
        End If
        ' -------------------------------------------------------- Calculate coefficients for the release segment
        '
        Dim d As Single = releaseTime * TimeScaler * SamplesPerSecond + 1
        Dim x As Single = CSng(Math.Exp(-1 / d))
        a0 = (1 - x) * 0.9F
        b1 = x
        m_state = State.Release
    End Sub


    Public ReadOnly Property HasCompleted() As Boolean
        Get
            Return m_state = State.Completed
        End Get
    End Property

    Public Overloads Overrides ReadOnly Property Ordinal() As Integer
        Get
            Return 1
        End Get
    End Property

    ' Gets or sets a value indicating whether the AdsrEnvelope overwrites its buffer when synthesizing its output.
    Public Overloads Overrides Property SynthesizeReplaceEnabled() As Boolean
        Get
            Return m_synthesizeReplaceEnabled
        End Get
        Set(ByVal value As Boolean)
            m_synthesizeReplaceEnabled = value
        End Set
    End Property



#Region "IProgrammable Members"

    Public Function GetParameterName(ByVal index As Integer) As String Implements IProgrammable.GetParameterName
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If
        '
        Dim result As String = ""
        Dim name__1 As String = Name
        If Not String.IsNullOrEmpty(name__1) Then name__1 = name__1 & " "
        '
        Select Case CType(index, ParameterId)
            Case ParameterId.AttackTime : result = name__1 & "Attack Time"
            Case ParameterId.DecayTime : result = name__1 & "Decay Time"
            Case ParameterId.SustainLevel : result = name__1 & "Sustain Level"
            Case ParameterId.ReleaseTime : result = name__1 & "Release Time"
            Case ParameterId.TriggerMode : result = name__1 & "Trigger Mode"
            Case ParameterId.VelocitySensitivity : result = name__1 & "Velocity Sensitivity"
            Case Else : Debug.Fail("Unhandled parameter")
        End Select
        Return result
    End Function

    Public Function GetParameterLabel(ByVal index As Integer) As String Implements IProgrammable.GetParameterLabel
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If
        Dim result As String = ""
        Select Case CType(index, ParameterId)
            Case ParameterId.AttackTime : result = "Time"
            Case ParameterId.DecayTime : result = "Time"
            Case ParameterId.SustainLevel : result = "dB"
            Case ParameterId.ReleaseTime : result = "Time"
            Case ParameterId.TriggerMode : result = "Retrigger/Legato"
            Case ParameterId.VelocitySensitivity : result = "Depth"
            Case Else : Debug.Fail("Unhandled parameter")
        End Select
        Return result
    End Function

    Public Function GetParameterDisplay(ByVal index As Integer) As String Implements IProgrammable.GetParameterDisplay
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If

        Dim result As String = String.Empty

        Select Case CType(index, ParameterId)
            Case ParameterId.AttackTime
                If True Then
                    Dim value As Integer = CInt(Math.Truncate(attackTime * ParameterScaler))

                    result = value.ToString(Synthesizer.FormatInfo)
                End If

            Case ParameterId.DecayTime
                If True Then
                    Dim value As Integer = CInt(Math.Truncate(decayTime * ParameterScaler))

                    result = value.ToString(Synthesizer.FormatInfo)
                End If

            Case ParameterId.SustainLevel
                result = Synthesizer.DecibelToString((sustainLevel - 1) / Ceiling)

            Case ParameterId.ReleaseTime
                If True Then
                    Dim value As Integer = CInt(Math.Truncate(releaseTime * ParameterScaler))

                    result = value.ToString(Synthesizer.FormatInfo)
                End If

            Case ParameterId.TriggerMode
                result = m_triggerMode.ToString()

            Case ParameterId.VelocitySensitivity
                If True Then
                    Dim value As Integer = CInt(Math.Truncate(velocitySensitivity * ParameterScaler))

                    result = value.ToString(Synthesizer.FormatInfo)
                End If

            Case Else
                Debug.Fail("Unhandled parameter")

        End Select

        Return result
    End Function

    Public Function GetParameterValue(ByVal index As Integer) As Single Implements IProgrammable.GetParameterValue
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If
        Dim result As Single = 0
        Select Case CType(index, ParameterId)
            Case ParameterId.AttackTime : result = attackTime
            Case ParameterId.DecayTime : result = decayTime
            Case ParameterId.SustainLevel : result = (sustainLevel - 1) / Ceiling
            Case ParameterId.ReleaseTime : result = releaseTime
            Case ParameterId.TriggerMode : result = (If(m_triggerMode = TriggerMode.Retrigger, 0, 1))
            Case ParameterId.VelocitySensitivity : result = velocitySensitivity
            Case Else : Debug.Fail("Unhandled parameter")
        End Select
        Return result
    End Function

    Public Sub SetParameterValue(ByVal index As Integer, ByVal value As Single) Implements IProgrammable.SetParameterValue
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        ElseIf value < 0 OrElse value > 1 Then
            Throw New ArgumentOutOfRangeException("value")
        End If
        Select Case CType(index, ParameterId)
            Case ParameterId.AttackTime : attackTime = value
            Case ParameterId.DecayTime : decayTime = value
            Case ParameterId.SustainLevel : sustainLevel = value * Ceiling + 1
            Case ParameterId.ReleaseTime : releaseTime = value
            Case ParameterId.TriggerMode : m_triggerMode = If(value < 0.5, TriggerMode.Retrigger, TriggerMode.Legato)
            Case ParameterId.VelocitySensitivity : velocitySensitivity = value
            Case Else : Debug.Fail("Unhandled parameter")
        End Select
    End Sub

    Public ReadOnly Property ParameterCount() As Integer Implements IProgrammable.ParameterCount
        Get
            Return ParameterId.VelocitySensitivity + 1
        End Get
    End Property

#End Region

End Class
