

Imports System.Collections.Generic
Imports System.Diagnostics
Imports System.IO
Imports System.Globalization
Imports System.Threading
Imports LC.Multimedia
Imports LC.Midi
Imports LC.Timers
Imports LC.Wave


Public Enum PolyMode
    Polyphonic
    Monophonic
End Enum


' ------------------------------------------------------------------------ represents a MIDI synthesizer
Partial Public NotInheritable Class Synthesizer
    Implements IProgrammable
    Implements IDisposable

    Public Const BufferSizeMinValue As Integer = 128
    Public Const BufferSizeMaxValue As Integer = 65536
    Public Const BufferSizeDefaultValue As Integer = 2048
    ' -------------------------------------------------------------------- the number of bytes per sample
    Friend Const BlockAlign As Integer = 4
    ' -------------------------------------------------------------------- The number of waveform buffers used
    Private Const BufferCount As Integer = 4
    ' -------------------------------------------------------------------- interval used by the waveform output device
    Private Const PollingInterval As Integer = 10
    ' -------------------------------------------------------------------- formating numbers displayed in the SynthControl Form.
    Friend Shared ReadOnly FormatInfo As NumberFormatInfo
    ' -------------------------------------------------------------------- for marshaling events to
    Private context As SynchronizationContext
    ' -------------------------------------------------------------------- indicates Synthesizer playing
    Private playing As Boolean = False
    ' -------------------------------------------------------------------- indicates Synthesizer recording
    Private recording As Boolean = False
    ' -------------------------------------------------------------------- for writing waveform data to a wave file
    Private waveWriter As WaveWriter
    ' -------------------------------------------------------------------- the name of the synthesizer
    Private synthName As String
    ' -------------------------------------------------------------------- the collection of voices the Synthesizer uses
    Private voices As New List(Of Voice)()
    ' -------------------------------------------------------------------- the raw waveform output from the Voices and effects
    Private synthBuffer As StereoBuffer
    ' -------------------------------------------------------------------- the rate at which the samples are sampled per second
    Private m_sampleRate As SampleRate
    ' -------------------------------------------------------------------- waveform device that plays the Synthesizer's output
    Private m_outputDevice As LC.Wave.OutputDevice
    ' -------------------------------------------------------------------- waveform buffers for holding the waveform output of the Synthesizer
    Private waveformBuffers As WaveformBuffer() = New WaveformBuffer(BufferCount - 1) {}
    ' -------------------------------------------------------------------- the effects used by the Synthesizer
    Private effects As List(Of EffectComponent)
    ' -------------------------------------------------------------------- allocating Voices in response to note-on-off events
    Private allocator As VoiceAllocator
    ' -------------------------------------------------------------------- bends pitch based on pitch bend wheel messages
    Private bender As PitchBender
    ' -------------------------------------------------------------------- processes controller messages
    Private controlProcessor As ControlProcessor
    ' -------------------------------------------------------------------- dispatches MIDI messages
    Private dispatcher As MessageDispatcher
    ' -------------------------------------------------------------------- holds all Synthesizer's IProgrammable objects
    Private parameters As New ProgrammableCollection()
    ' -------------------------------------------------------------------- the Synthesizer's bank of programs ("patches")
    Private bank As Bank
    ' -------------------------------------------------------------------- the currently selected program index
    Private programIndex As Integer = 0
    ' -------------------------------------------------------------------- for queueing synthesizer events
    Private eventQueue As New Queue(Of SynthEvent)()
    ' -------------------------------------------------------------------- for thread safe operations
    Private ReadOnly lockObject As New Object()
    ' -------------------------------------------------------------------- indicates the Synthesizer has been disposed
    Private disposed As Boolean = False


    Public Event ParameterChanged As EventHandler(Of ParameterChangedEventArgs)
    Public Event SelectedProgramIndexChanged As EventHandler
    Public Event PlayingStarted As EventHandler
    Public Event PlayingStopped As EventHandler
    Public Event BankLoaded As EventHandler
    Public Event ProgramLoaded As EventHandler
    Public Event [Error] As EventHandler(Of LC.Multimedia.ErrorEventArgs)
    Public Event BufferPlayed As EventHandler(Of BufferPlayedEventArgs)


    Shared Sub New()
        FormatInfo = New NumberFormatInfo()
        FormatInfo.NumberDecimalDigits = 2
    End Sub


    Public Sub New(ByVal synthName As String, _
                   ByVal deviceId As Integer, _
                   ByVal bufferSize As Integer, _
                   ByVal sampleRate As Integer, _
                   ByVal voiceFactory As VoiceFactory, _
                   ByVal voiceCount As Integer)

        ' Dummy effect factory delegate.
        Initialize(deviceId, _
                   bufferSize, _
                   sampleRate, _
                   voiceFactory, _
                   voiceCount, _
                   synthName, _
                   Function(sr As SampleRate, inputBuffer As StereoBuffer) New EffectComponent(-1) {})
    End Sub


    Public Sub New(ByVal synthName As String, _
                   ByVal deviceId As Integer, _
                   ByVal bufferSize As Integer, _
                   ByVal sampleRate As Integer, _
                   ByVal voiceFactory As VoiceFactory, _
                   ByVal voiceCount As Integer, _
                   ByVal effectFactory As EffectFactory)

        Initialize(deviceId, _
                   bufferSize, _
                   sampleRate, _
                   voiceFactory, _
                   voiceCount, _
                   synthName, _
                   effectFactory)
    End Sub




    Private Sub Initialize(ByVal deviceId As Integer, _
                           ByVal bufferSize As Integer, _
                           ByVal sampleRate As Integer, _
                           ByVal voiceFactory As VoiceFactory, _
                           ByVal voiceCount As Integer, _
                           ByVal synthName As String, _
                           ByVal effectFactory As EffectFactory)

        If synthName Is Nothing Then
            Throw New ArgumentNullException("synthName")
        ElseIf bufferSize < BufferSizeMinValue OrElse bufferSize > BufferSizeMaxValue Then
            Throw New ArgumentOutOfRangeException("bufferSize")
        ElseIf sampleRate < 1 Then
            Throw New ArgumentOutOfRangeException("sampleRate")
        ElseIf voiceFactory Is Nothing Then
            Throw New ArgumentNullException("voiceFactory")
        ElseIf voiceCount < 1 Then
            Throw New ArgumentOutOfRangeException("voiceCount")
        ElseIf effectFactory Is Nothing Then
            Throw New ArgumentNullException("effectFactory")
        End If

        ' Create waveform output device here first so that if an exception
        ' is thrown, there's no need for clean up.
        m_outputDevice = New LC.Wave.OutputDevice(deviceId, 2, sampleRate, 16)
        If Not m_outputDevice.IsWorking Then
            m_outputDevice = Nothing
            Return
        End If

        m_outputDevice.PollingInterval = PollingInterval

        Me.m_sampleRate = New SampleRate(sampleRate)
        Me.synthName = synthName

        synthBuffer = New StereoBuffer(bufferSize)

        CreateVoices(voiceFactory, voiceCount)

        allocator = New VoiceAllocator(voices)
        bender = New PitchBender(voices)
        controlProcessor = New ControlProcessor(voices, allocator)
        dispatcher = New MessageDispatcher(allocator, controlProcessor, bender)

        parameters.Add(synthBuffer)
        parameters.Add(bender)
        parameters.Add(allocator)
        parameters.Add(New VoiceParameters(voices))

        ' Create effects.
        Me.effects = New List(Of EffectComponent)(effectFactory(Me.m_sampleRate, synthBuffer))

        ' For each effect that implements IProgrammable, add it to the 
        ' collection of IProgrammable objects.
        For Each effect As EffectComponent In effects
            Dim programmableEffect As IProgrammable = TryCast(effect, IProgrammable)

            If programmableEffect IsNot Nothing Then
                parameters.Add(programmableEffect)
            End If
        Next

        bank = New Bank(synthName, ParameterCount)

        InitializeBank()

        For i As Integer = 0 To waveformBuffers.Length - 1
            waveformBuffers(i) = New WaveformBuffer(bufferSize * BlockAlign)
        Next

        '
        ' Get/Create SynchronizationContext object for marshaling events.
        '
        If SynchronizationContext.Current Is Nothing Then
            context = New SynchronizationContext()
        Else
            context = SynchronizationContext.Current
        End If
    End Sub

    Private Sub CreateVoices(ByVal factory As VoiceFactory, ByVal voiceCount As Integer)
        Debug.Assert(voiceCount >= 1)

        For i As Integer = 0 To voiceCount - 1
            voices.Add(factory(m_sampleRate, synthBuffer, String.Empty))
            voices(i).SynthesizeReplaceEnabled = False
        Next
    End Sub

    Private Sub InitializeBank()
        For i As Integer = 0 To bank.ProgramCount - 1
            For p As Integer = 0 To ParameterCount - 1
                bank(i)(p) = GetParameterValue(p)
            Next
        Next
    End Sub

    Private Sub WaveformBufferCallback(ByVal result As IAsyncResult)
        If playing Then
            Try
                m_outputDevice.EndSend(result)

                Dim buffer As WaveformBuffer = DirectCast(result.AsyncState, WaveformBuffer)

                SyncLock lockObject
                    WriteToBuffer(buffer)

                    m_outputDevice.BeginSend(buffer, AddressOf WaveformBufferCallback, buffer)
                End SyncLock
            Catch ex As LC.Wave.OutputDeviceException
                OnError(New LC.Multimedia.ErrorEventArgs(ex))
            End Try
        End If
    End Sub


    Public Sub PlayStart()
        If playing Then Return
        '
        SyncLock lockObject
            '
            playing = True
            synthBuffer.Clear()
            ResetEffects()
            '
            For Each buffer As WaveformBuffer In waveformBuffers
                buffer.Clear()
                m_outputDevice.BeginSend(buffer, AddressOf WaveformBufferCallback, buffer)
            Next
        End SyncLock
        '
        RaiseEvent PlayingStarted(Me, EventArgs.Empty)
    End Sub


    Public Sub PlayStop()
        If Not playing Then Return
        '
        SyncLock lockObject
            playing = False
            eventQueue.Clear()
            m_outputDevice.Reset()
        End SyncLock
        '
        RaiseEvent PlayingStopped(Me, EventArgs.Empty)
    End Sub


    Public Sub StartRecording(ByVal output As Stream)
        If recording Then StopRecording()
        '
        SyncLock lockObject
            waveWriter = New WaveWriter(output, 2, m_sampleRate.SamplesPerSecond, 16)
            recording = True
        End SyncLock
    End Sub


    Public Sub StopRecording()
        If Not recording Then Return
        '
        SyncLock lockObject
            waveWriter.Close()
            recording = False
        End SyncLock
    End Sub

    ' -------------------------------------------------------------------- Processes MIDI channel messages
    Public Sub ProcessChannelMessage(ByVal message As ChannelMessage)
        If m_outputDevice Is Nothing Then Return
        ' ---------------------------------------------------------------- Get position in samples
        Dim t As Time = m_outputDevice.GetPosition(TimeType.Samples)
        Debug.Assert(t.type = TimeType.Samples)
        ' ---------------------------------------------------------------- Calculate delta timestamp in samples
        Dim deltaSamples As Integer = t.samples Mod synthBuffer.BufferSize
        '
        SyncLock eventQueue
            ' ------------------------------------------------------------- Enqueue event
            eventQueue.Enqueue(New SynthEvent(deltaSamples, message))
        End SyncLock
    End Sub


    Public Sub SetSelectedProgramIndex(ByVal index As Integer)
        If index < 0 OrElse index >= ProgramCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If
        If programIndex = index Then Return

        SyncLock lockObject
            Dim wasPlaying As Boolean = playing
            PlayStop()
            programIndex = index
            ResetProgram()
            RaiseEvent SelectedProgramIndexChanged(Me, EventArgs.Empty)
            If wasPlaying Then PlayStart()
        End SyncLock
    End Sub


    Public Sub SetBufferSize(ByVal bufferSize As Integer)
        If bufferSize < BufferSizeMinValue OrElse bufferSize > BufferSizeMaxValue Then
            Throw New ArgumentOutOfRangeException("bufferSize")
        End If
        If bufferSize = synthBuffer.BufferSize Then Return

        SyncLock lockObject
            Dim wasPlaying As Boolean = playing
            PlayStop()
            synthBuffer.BufferSize = bufferSize
            '
            For i As Integer = 0 To BufferCount - 1
                waveformBuffers(i).Dispose()
                waveformBuffers(i) = New WaveformBuffer(bufferSize * BlockAlign)
            Next
            '
            If wasPlaying Then PlayStart()
        End SyncLock
    End Sub

    Public Sub SetSampleRate(ByVal sampleRate As Integer)
        If sampleRate < 1 Then
            Throw New ArgumentOutOfRangeException("sampleRate")
        End If

        SyncLock lockObject
            Dim wasPlaying As Boolean = playing
            PlayStop()
            Dim deviceId As Integer = m_outputDevice.DeviceID
            '
            m_outputDevice.Dispose()
            m_outputDevice = New LC.Wave.OutputDevice(deviceId, 2, sampleRate, 16)
            m_outputDevice.PollingInterval = PollingInterval
            Me.m_sampleRate.SamplesPerSecond = sampleRate
            '
            If wasPlaying Then PlayStart()
        End SyncLock
    End Sub


    Public Sub SetWaveDevice(ByVal deviceId As Integer)
        If deviceId = m_outputDevice.DeviceID Then
            Return
        End If
        '
        SyncLock lockObject
            Dim wasPlaying As Boolean = playing
            Try
                PlayStop()
                Dim newOutputDevice As New LC.Wave.OutputDevice(deviceId, 2, m_sampleRate.SamplesPerSecond, 16)
                '
                newOutputDevice.PollingInterval = PollingInterval
                m_outputDevice.Dispose()
                m_outputDevice = newOutputDevice
            Finally
                If wasPlaying Then PlayStart()
            End Try
        End SyncLock
    End Sub


    ' --------------------------------------------------------- gets the Program name at the ( zero-based ) index
    Public Function GetProgramName(ByVal index As Integer) As String
        If index < 0 OrElse index >= ProgramCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If
        Return bank(index).Name
    End Function

    Private Sub ResetProgram()
        For i As Integer = 0 To ParameterCount - 1
            SetParameterValueHelper(i, bank(programIndex)(i))
        Next
    End Sub

    Private Sub ResetEffects()
        For Each effect As EffectComponent In effects
            effect.Reset()
        Next
    End Sub

    Private Sub OnBufferPlayed(ByVal e As BufferPlayedEventArgs)
        'Dim handler As EventHandler(Of BufferPlayedEventArgs) = BufferPlayed
        'If handler IsNot Nothing Then
        '	context.Post(New SendOrPostCallback(Function(dummy As Object) Do
        '		handler(Me, e)
        '	End Function), Nothing)
        'End If
        context.Post(Function(dummy As Object) BufferPlayed_Function(e), Nothing)
    End Sub
    Private Function BufferPlayed_Function(ByVal e As BufferPlayedEventArgs) As Boolean
        RaiseEvent BufferPlayed(Me, e)
    End Function

    Private Sub OnError(ByVal e As LC.Multimedia.ErrorEventArgs)
        'Dim handler As EventHandler(Of ErrorEventArgs) = [Error]
        'If handler IsNot Nothing Then
        '	context.Post(New SendOrPostCallback(Function(dummy As Object) Do
        '		handler(Me, e)
        '	End Function), Nothing)
        'End If
        context.Post(Function(dummy As Object) OnError_Function(e), Nothing)
    End Sub
    Private Function OnError_Function(ByVal e As LC.Multimedia.ErrorEventArgs) As Boolean
        RaiseEvent [Error](Me, e)
    End Function


    Private Sub WriteToBuffer(ByVal buffer As WaveformBuffer)
        Dim deltaSamples As Integer = 0
        Dim offset As Integer = 0
        Dim count As Integer
        Dim synthEvent As SynthEvent
        '
        synthBuffer.Clear()
        ' ---------------------------------------------------------------- while the buffer has not been filled
        While offset < synthBuffer.BufferSize
            SyncLock eventQueue
                ' -------------------------------------------------------- there is an event in the queue 
                '                                                          and it comes after the current position.
                If eventQueue.Count > 0 AndAlso _
                    eventQueue.Peek().DeltaSamples >= deltaSamples Then
                    ' ---------------------------------------------------- dequeue event from queue
                    synthEvent = eventQueue.Dequeue()
                    ' ---------------------------------------------------- update position
                    deltaSamples = synthEvent.DeltaSamples
                    ' ---------------------------------------------------- calculate number of samples to synthesize
                    count = deltaSamples - offset + 1
                Else
                    ' ---------------------------------------------------- there are no events or if there are any,
                    '                                                       they are before the current position
                    '
                    ' ---------------------------------------------------- indicate that no event was dequeued
                    synthEvent = Nothing

                    ' ---------------------------------------------------- calculate number of samples to synthesize
                    count = synthBuffer.BufferSize - offset
                End If
            End SyncLock

            ' ------------------------------------------------------------ if there are samples to synthesize
            If count > 0 Then
                ' -------------------------------------------------------- synthesize each Voice's contribution to mix
                For Each voice As Voice In voices
                    voice.Synthesize(offset, count)
                Next
                ' -------------------------------------------------------- move offset forward
                offset += count
            End If
            ' ------------------------------------------------------------ if there is an event to process
            If synthEvent IsNot Nothing Then
                ' -------------------------------------------------------- dispatch message
                dispatcher.Dispatch(synthEvent.Message)
            End If
        End While

        ' ---------------------------------------------------------------- run each effect
        For Each effect As EffectComponent In effects
            effect.Process()
        Next

        synthBuffer.ModulateAmplitude()

        synthBuffer.WriteToOutputBuffer(buffer)

        OnBufferPlayed(New BufferPlayedEventArgs(synthBuffer.GetLeftPeak(), synthBuffer.GetRightPeak()))

        ' ---------------------------------------------------------------- if this synth is recording
        If recording Then
            waveWriter.Write(buffer.GetBuffer())
        End If
    End Sub

    Private Sub SetParameterValueHelper(ByVal index As Integer, ByVal value As Single)
        SyncLock lockObject
            Debug.Assert(index < parameters.ParameterCount)
            parameters.SetParameterValue(index, value)
            ' ------------------------------------------------------------ keep track of raw parameter value
            bank(SelectedProgramIndex)(index) = value
        End SyncLock
    End Sub


    Public ReadOnly Property SelectedProgramIndex() As Integer
        Get
            Return programIndex
        End Get
    End Property

    Public Property SelectedProgramName() As String
        Get
            Return bank(programIndex).Name
        End Get
        Set(ByVal value As String)
            bank(programIndex).Name = value
        End Set
    End Property

    Public Property MidiChannel() As Integer
        Get
            Return dispatcher.MidiChannel
        End Get
        Set(ByVal value As Integer)
            dispatcher.MidiChannel = value
        End Set
    End Property

    Public Property OmniEnabled() As Boolean
        Get
            Return dispatcher.OmniEnabled
        End Get
        Set(ByVal value As Boolean)
            dispatcher.OmniEnabled = value
        End Set
    End Property

    Public ReadOnly Property BufferSize() As Integer
        Get
            Return synthBuffer.BufferSize
        End Get
    End Property

    Public ReadOnly Property SampleRate() As Integer
        Get
            Return m_sampleRate.SamplesPerSecond
        End Get
    End Property

    Public ReadOnly Property ProgramCount() As Integer
        Get
            Return bank.ProgramCount
        End Get
    End Property

    Public ReadOnly Property Name() As String
        Get
            Return synthName
        End Get
    End Property

    Public ReadOnly Property IsPlaying() As Boolean
        Get
            Return playing
        End Get
    End Property



#Region "IDisposable Members"

    Public Sub Dispose() Implements IDisposable.Dispose
        If disposed Then Return
        '
        PlayStop()
        StopRecording()
        disposed = True
        m_outputDevice.Reset()
        For Each buffer As WaveformBuffer In waveformBuffers
            buffer.Dispose()
        Next
        m_outputDevice.Dispose()
    End Sub

#End Region

#Region "IProgrammable Members"

    Public Function GetParameterName(ByVal index As Integer) As String Implements IProgrammable.GetParameterName
        Return parameters.GetParameterName(index)
    End Function

    Public Function GetParameterLabel(ByVal index As Integer) As String Implements IProgrammable.GetParameterLabel
        Return parameters.GetParameterLabel(index)
    End Function

    Public Function GetParameterDisplay(ByVal index As Integer) As String Implements IProgrammable.GetParameterDisplay
        Return parameters.GetParameterDisplay(index)
    End Function

    Public Function GetParameterValue(ByVal index As Integer) As Single Implements IProgrammable.GetParameterValue
        Return parameters.GetParameterValue(index)
    End Function

    Public Sub SetParameterValue(ByVal index As Integer, ByVal value As Single) Implements IProgrammable.SetParameterValue
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        ElseIf value < 0 OrElse value > 1 Then
            Throw New ArgumentOutOfRangeException("value")
        End If

        If GetParameterValue(index) = value Then Return

        SetParameterValueHelper(index, value)

        RaiseEvent ParameterChanged(Me, New ParameterChangedEventArgs(index, value))
    End Sub

    Public ReadOnly Property ParameterCount() As Integer Implements IProgrammable.ParameterCount
        Get
            Return parameters.ParameterCount
        End Get
    End Property

#End Region

End Class
