

Imports System.Diagnostics
Imports System.IO

Public NotInheritable Class WaveReader
	Implements IDisposable

    Private Const ChunkIDLength As Integer = 4

    Private input As Stream
    Private m_channelCount As Short
    Private m_samplesPerSecond As Integer
    Private m_bitsPerSecond As Short
    Private waveChunkPosition As Long
    Private bytesLeft As Long
    Private m_waveDataLength As Long
    Private disposed As Boolean = False

    Public Sub New(ByVal input As Stream)
        If input Is Nothing Then
            Throw New ArgumentNullException("input")
        ElseIf Not input.CanRead Then
            Throw New ArgumentException("The input stream cannot be read.")
        ElseIf Not input.CanSeek Then
            Throw New ArgumentException("The input stream does not support seeking.")
        End If

        Me.input = input

        input.Position = 0

        If IsRiffFile() Then
            ' --------------------------------------------------------------- skip RIFF length
            input.ReadByte()
            input.ReadByte()
            input.ReadByte()
            input.ReadByte()

            waveChunkPosition = FindChunk(New Byte() {Asc("W"c), Asc("A"c), Asc("V"c), Asc("E"c)})

            If waveChunkPosition < 0 Then
                Throw New ChunkNotFoundException("WAVE")
            End If

            If FindChunk(New Byte() {Asc("f"c), Asc("m"c), Asc("t"c), Asc(" "c)}) < 0 Then
                Throw New ChunkNotFoundException("fmt ")
            End If

            ParseFormatData()

            ' ------------------------------------------------ reposition stream to the beginning of the wave chunk
            input.Position = waveChunkPosition

            If FindChunk(New Byte() {Asc("d"c), Asc("a"c), Asc("t"c), Asc("a"c)}) < 0 Then
                Throw New ChunkNotFoundException("data")
            End If

            ParseWaveData()
        Else
            Throw New ChunkNotFoundException("RIFF")
        End If
    End Sub


    Private Function IsRiffFile() As Boolean
        Dim result As Boolean
        If input.ReadByte() = Asc("R"c) AndAlso input.ReadByte() = Asc("I"c) AndAlso input.ReadByte() = Asc("F"c) AndAlso input.ReadByte() = Asc("F"c) Then
            result = True
        Else
            result = False
        End If
        Return result
    End Function

    Private Function FindChunk(ByVal chunkID As Byte()) As Long
        Debug.Assert(chunkID IsNot Nothing AndAlso chunkID.Length = ChunkIDLength)

        Dim found As Boolean = False
        Dim b As Integer
        Dim i As Integer = 0
        ' ------------------------------------------------------------- "Prime the pump."
        b = input.ReadByte()

        While Not found AndAlso b >= 0
            If b = chunkID(i) Then
                i += 1

                If i < chunkID.Length Then
                    b = input.ReadByte()
                Else
                    found = True
                End If
            Else
                b = input.ReadByte()
                i = 0
            End If
        End While

        Return (If(found, input.Position, -1))
    End Function

    Private Sub ParseFormatData()
        ' -------------------------------------------------------------- skip format length
        input.ReadByte()
        input.ReadByte()
        input.ReadByte()
        input.ReadByte()

        Dim shortArray As Byte() = New Byte(1) {}
        Dim intArray As Byte() = New Byte(3) {}

        input.Read(shortArray, 0, shortArray.Length)
        Dim formatTag As Short = BitConverter.ToInt16(shortArray, 0)

        If formatTag <> 1 Then
            ' If the format is not PCM/Uncompressed, throw an exception.
            Throw New NotSupportedException("Non PCM/Uncompressed formats are not supported at this time.")
        End If

        input.Read(shortArray, 0, shortArray.Length)
        m_channelCount = BitConverter.ToInt16(shortArray, 0)

        input.Read(intArray, 0, intArray.Length)
        m_samplesPerSecond = BitConverter.ToInt32(intArray, 0)

        ' ------------------------------------------------------------------ skip average bytes per second
        input.Read(intArray, 0, intArray.Length)
        ' ------------------------------------------------------------------ skip block align
        input.Read(shortArray, 0, shortArray.Length)
        '
        input.Read(shortArray, 0, shortArray.Length)
        m_bitsPerSecond = BitConverter.ToInt16(shortArray, 0)
    End Sub

    Private Sub ParseWaveData()
        Dim uintArray As Byte() = New Byte(3) {}
        input.Read(uintArray, 0, uintArray.Length)
        bytesLeft = BitConverter.ToUInt32(uintArray, 0)
        m_waveDataLength = bytesLeft
    End Sub


    Public Function Read() As Byte
        If disposed Then
            Throw New ObjectDisposedException(Me.[GetType]().Name)
        ElseIf bytesLeft = 0 Then
            Throw New EndOfStreamException()
        End If

        bytesLeft -= 1

        Dim b As Integer = input.ReadByte()

        Debug.Assert(b >= 0)

        Return CByte(b)
    End Function

    Public Function Read(ByVal count As Integer) As Byte()
        If disposed Then
            Throw New ObjectDisposedException(Me.[GetType]().Name)
        ElseIf count < 0 Then
            Throw New ArgumentOutOfRangeException("count")
        End If

        Dim buffer As Byte()

        If count <= bytesLeft Then
            buffer = New Byte(count - 1) {}

            bytesLeft -= count
        Else
            buffer = New Byte(CInt(bytesLeft) - 1) {}

            bytesLeft = 0
        End If

        input.Read(buffer, 0, buffer.Length)

        Return buffer
    End Function

    Public Sub Close()
        Dispose()
    End Sub


    Public ReadOnly Property BaseStream() As Stream
        Get
            Return input
        End Get
    End Property

    Public ReadOnly Property ChannelCount() As Short
        Get
            Return m_channelCount
        End Get
    End Property

    Public ReadOnly Property SamplesPerSecond() As Integer
        Get
            Return m_samplesPerSecond
        End Get
    End Property

    Public ReadOnly Property BitsPerSecond() As Short
        Get
            Return m_bitsPerSecond
        End Get
    End Property

    Public ReadOnly Property WaveDataLength() As Long
        Get
            Return m_waveDataLength
        End Get
    End Property


    Public Sub Dispose() Implements IDisposable.Dispose
        If disposed Then
            Return
        End If
        input.Close()
        disposed = True
        GC.SuppressFinalize(Me)
    End Sub

End Class
